/* Part of Misnamed Filter Configurator (see Program.cs)
 * by Alex Vallat <www.byalexv.co.uk>
 */
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Text.RegularExpressions;
using System.Reflection;

namespace MisnamedFilter.Configurator
{
	public partial class Tester : Form
	{
		private Main mParent;
		private MethodInfo mDecodeHeaderValue;

		public Tester()
		{
			InitializeComponent();

			//Obtain the internal System.Net.Mime.MimeBasePart.DecodeHeaderValue method.
			//This is not publically exposed, so have to guard against it disappearing in future CLR versions.
			try
			{
				Type mimeBasePart = typeof(System.Net.Mime.TransferEncoding).Assembly.GetType("System.Net.Mime.MimeBasePart");
				if (mimeBasePart != null)
				{
					mDecodeHeaderValue = mimeBasePart.GetMethod("DecodeHeaderValue", BindingFlags.Static | BindingFlags.NonPublic);
				}
			}
			catch (Exception)
			{
				//Couldn't find the DecodeHeaderValue method
			}
		}
		internal Tester(Main parent)
			: this()
		{
			mParent = parent;
			Font = parent.Font;
		}

		protected override void OnClosing(CancelEventArgs e)
		{
			base.OnClosing(e);
			AbortTest();
		}

		private void mClose_Click(object sender, EventArgs e)
		{
			this.Close();
		}

		private void mDisplayName_Enter(object sender, EventArgs e)
		{
			UpdateDisplayNameToolTip();
		}
		private void mDisplayName_Leave(object sender, EventArgs e)
		{
			mToolTip.Hide(mDisplayName);
		}

		private void UpdateDisplayNameToolTip()
		{
			if (mToolTip.Site != null) //Will be null if form is closing, and it has been disconnected
			{
				string decodedDisplayName = DecodeHeaderValue(mDisplayName.Text);
				if (decodedDisplayName.Equals(mDisplayName.Text, StringComparison.OrdinalIgnoreCase))
				{
					mToolTip.Hide(mDisplayName);
				}
				else
				{
					mToolTip.Show(decodedDisplayName, mDisplayName, 0, mDisplayName.Height);
				}
			}
		}

		private void mEmailAddress_TextChanged(object sender, EventArgs e)
		{
			BeginTest();
		}

		private void mDisplayName_TextChanged(object sender, EventArgs e)
		{
			UpdateDisplayNameToolTip();
			BeginTest();
		}

		/// <summary>
		/// Asynchronously starts testing, cancelling any existing test
		/// </summary>
		internal void BeginTest()
		{
			AbortTest();

			TestParameters parameters = new TestParameters();

			parameters.EmailAddress = mEmailAddress.Text;
			parameters.DisplayName = mDisplayName.Text;

			if (String.IsNullOrEmpty(parameters.EmailAddress) ||
				String.IsNullOrEmpty(parameters.DisplayName) )
			{
				//Don't bother testing empty name or email
				return;
			}

			//Get the current set of address sets
			parameters.AddressSets = new List<AddressSet>(mParent.GetAddressSets());

			//Start the worker for doing the test
			lock (mTestRunning) //Using the lock to prevent wait blocking while invoking adding a match result
			{
				mTestRunning.WaitOne(); //Wait until the currently running test finishes.
			}
			mAbortTest = false;
			mResults.Nodes.Clear();
			ThreadPool.QueueUserWorkItem(new WaitCallback(PerformTest), parameters);
		}

		/// <summary>
		/// Aborts an asynchronously started test
		/// </summary>
		private void AbortTest()
		{
			mAbortTest = true;
		}

		private struct TestParameters
		{
			public List<AddressSet> AddressSets;
			public string EmailAddress;
			public string DisplayName;
		}

		private delegate void MatchFound(string addressSetName, string emailAddressPattern, string namePattern);
		/// <summary>Test is running while non-signalled.</summary>
		private ManualResetEvent mTestRunning = new ManualResetEvent(true);
		private volatile bool mAbortTest = false;
		private void PerformTest(object state)
		{
			mTestRunning.Reset();
			TestParameters parameters = (TestParameters)state;
			
			foreach (AddressSet addressSet in parameters.AddressSets)
			{
				foreach (String candidateAddressPattern in addressSet.AddressPatterns)
				{
					if (!String.IsNullOrEmpty(candidateAddressPattern))
					{
						try
						{
							//Treat the address as a regex and check it against the address
							if (Regex.IsMatch(parameters.EmailAddress, candidateAddressPattern, RegexOptions.IgnoreCase | RegexOptions.ExplicitCapture))
							{
								lock (mTestRunning)
								{
									Invoke(new MatchFound(OnMatchFound), addressSet.DescriptiveName, candidateAddressPattern, null);
								}
								//Found a matching address which needs checking, so check that the name is contained within the display name
								foreach (String candidateNamePattern in addressSet.NamePatterns)
								{
									if (!String.IsNullOrEmpty(candidateNamePattern))
									{
										try
										{
											//Treat the name as a regex too, checking first undecoded, then decoded value
											if (Regex.IsMatch(parameters.DisplayName, candidateNamePattern, RegexOptions.IgnoreCase | RegexOptions.ExplicitCapture) ||
												Regex.IsMatch(DecodeHeaderValue(parameters.DisplayName), candidateNamePattern, RegexOptions.IgnoreCase | RegexOptions.ExplicitCapture))
											{
												//This address is directed to a valid name.
												lock (mTestRunning)
												{
													Invoke(new MatchFound(OnMatchFound), addressSet.DescriptiveName, candidateAddressPattern, candidateNamePattern);
												}
											} // </Is the name a match>
										}
										catch (ArgumentException e)
										{
											lock (mTestRunning)
											{
												Invoke(new MatchFound(OnMatchFound), addressSet.DescriptiveName, candidateAddressPattern, String.Format("! Regex Error in Name Pattern \"{0}\": {1}", candidateNamePattern, e.Message));
											}
										}
									}
									if (mAbortTest) break;
								} // </For each namePattern>
							} // </Is the address a match>
						}
						catch (ArgumentException e)
						{
							lock (mTestRunning)
							{
								Invoke(new MatchFound(OnMatchFound), addressSet.DescriptiveName, String.Format("! Regex Error in Address Pattern \"{0}\": {1}", candidateAddressPattern, e.Message), null);
							}
						}
					}
					if (mAbortTest) break;
				}// </For each candidateAddressPatern>
				if (mAbortTest) break;
			} // </For each address set>
			mTestRunning.Set();
		}
		private void OnMatchFound(string addressSetName, string emailAddressPattern, string namePattern)
		{
			if (!mResults.Nodes.ContainsKey(addressSetName))
				mResults.Nodes.Add(addressSetName, addressSetName);

			TreeNode addressSetNode = mResults.Nodes[addressSetName];

			TreeNode emailAddressNode = addressSetNode.Nodes[emailAddressPattern];
			if (emailAddressNode == null)
			{
				emailAddressNode = addressSetNode.Nodes.Add(emailAddressPattern, emailAddressPattern);
				emailAddressNode.ForeColor = Color.Red;
				addressSetNode.Expand();
			}

			if (namePattern != null)
			{
				emailAddressNode.ForeColor = mResults.ForeColor;
				if (!emailAddressNode.Nodes.ContainsKey(namePattern))
				{
					emailAddressNode.Nodes.Add(namePattern);
					emailAddressNode.Expand();
				}
			}

			UpdateSummary();
		}

		private void UpdateSummary()
		{
			foreach (TreeNode addressSetNode in mResults.Nodes)
			{
				foreach (TreeNode emailAddressNode in addressSetNode.Nodes)
				{
					if (emailAddressNode.Nodes.Count == 0)
					{
						//Spam summary
						mResultSummary.Text = "This email address and name WOULD indicate spam, as there are address patterns which match which have no name paterns that match (indicated in red).";
						return;
					}
				}
			}
			//No spam summary
			mResultSummary.Text =             "This email address and name would NOT indicate spam, because all address sets which have matching address patterns also have name patterns that match.";
		}


		private string DecodeHeaderValue(string value)
		{
			//Call the internal CLR DecodeHeaderValue method.
			//This is not publically exposed, so have to guard against it disappearing in future CLR versions.
			if (mDecodeHeaderValue != null)
			{
				try
				{
					return (string)mDecodeHeaderValue.Invoke(null, new object[] { value });
				}
				catch (Exception)
				{
					//Couldn't decode the header (something went wrong)
				}
			}
			return value;
		}
	}
}