/* Misnamed Filter
 * by Alex Vallat <www.byalexv.co.uk>
 *
 * Detects emails sent to your email address, but someone elses name.
 *
 * 2007-02-03: v0.1
 *  Initial Release (AV)
 *
 * Some code copied from Spamihilator SDK exampleplugin.cpp
 *
 * Source code free for non-commercial use
 * Please share any improvements or changes
*/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "plugins_services.h"
#include "plugins_spamcheck.h"
#include "Settings.h"

using namespace System;
using namespace System::IO;
using namespace System::Collections::Generic;
using namespace System::Text::RegularExpressions;
using namespace System::Globalization;
using namespace System::Windows::Forms;
using namespace System::Reflection;
using namespace MisnamedFilter;

/**
 * Receiver identifier - this constant is reserved for future releases of
 * the Plugin SDK.
 */
static const char RECVID_MISNAMEDFILTER[] = "Misnamed Filter";

/**
 * Pointer to Spamihilator's SendService function
 * The plugin can send services to Spamihilator through this function
 */
static LPSENDSERVICEFUNC SendService = NULL;

ref class GlobalObjects {
public:
   static Regex^ EmailParser;
   static Regex^ HeaderParser;
   static Settings^ Settings;
   static array<Char>^ DisplayNameTrimChars = {' ', '\n', '"', '\''};
   static MethodInfo^ DecodeHeaderValue;
}; 

/**
 * Forward declarations
 */
int isSpamHdr(unsigned int id, const MSGHDR *header, double *probability);
int init();
void config(HWND parent);
long onLearn(short flag, const char *messageText);

extern "C" __declspec(dllexport) long ReceiveService(SERVICE *service)
{
	switch (service->ulService)
	{
	case SERV_INIT:
		SendService = (LPSENDSERVICEFUNC)service->lParam;
		break;

	case SERV_GETRECV:
		return (long)RECVID_MISNAMEDFILTER;

	case SERV_LEARN:
		//Spamihilator wants us to learn something from a special message
		return onLearn((short)service->wParam, (const char *)service->lParam);
	}
	return 0;
}

/**
 * SPAMCHECKPLUGININFO structure containing informaion about the plugin
 */
SPAMCHECKPLUGININFO spamcheckplugininfo = {
	MAKEVERSION(0,5,0),   //plugin version (0.1.0)
	MAKEVERSION(0,9,8),   //Spamihilator version needed (0.9.8)
	"Misnamed Filter",     //plugin name
	"Alex Vallat",      //author's name
	"Detects emails sent to your email address, but someone elses name", //plugin description

	init,				  //init()
	0,                    //free()
	config,				  //config()

	isSpamHdr,            //isSpamHdr()
	0,                    //isSpamHtml()
	0,                    //isSpamNoHtml()
	0                     //isSpam()
};

/**
 * Returns a pointer to a SPAMCHECKPLUGININFO structure containing
 * information about the plugin.
 *
 * @return Pointer to a SPAMCHECKPLUGININFO structure
 */
extern "C" __declspec(dllexport) SPAMCHECKPLUGININFO *GetSpamCheckPluginInfo()
{
	return &spamcheckplugininfo;
}

int init()
{
	GlobalObjects::EmailParser = gcnew Regex("(,\\s*)*(?<name>(?:[^,]+[^\"',]|\"[^\"]+\"|'[^']+'))\\s+<(?<address>[^>]+)>",
											RegexOptions::ExplicitCapture | RegexOptions::Compiled);

	GlobalObjects::HeaderParser = gcnew Regex("^(to|b?cc): ?(?<header>.*(\\n\\s.*)*)$", 
											RegexOptions::ExplicitCapture | RegexOptions::Multiline | RegexOptions::IgnoreCase | RegexOptions::Compiled);
	GlobalObjects::Settings = gcnew Settings();

	//Obtain the internal System.Net.Mime.MimeBasePart.DecodeHeaderValue method.
	//This is not publically exposed, so have to guard against it disappearing in future CLR versions.
	try
	{
		Type^ mimeBasePart = System::Net::Mime::TransferEncoding::typeid->Assembly->GetType("System.Net.Mime.MimeBasePart");
		if (mimeBasePart != nullptr)
		{
			GlobalObjects::DecodeHeaderValue = mimeBasePart->GetMethod("DecodeHeaderValue", BindingFlags::Static | BindingFlags::NonPublic);
		}
	}
	catch (Exception^)
	{
		//Couldn't find the DecodeHeaderValue method
	}
	
	return 0;
}

void config(HWND parent)
{
	//Try to find the Misnamed Filter folder
	DirectoryInfo^ settingsFolder = gcnew DirectoryInfo(Settings::GetSettingsFolder());
	if(!settingsFolder->Exists)
	{
		try
		{
			settingsFolder->Create();
		}
		catch(Exception^ e)
		{
			MessageBox::Show(String::Format("Could not create settings folder folder:\n\n {0}\n\n{1}", settingsFolder, e->Message));
			return;
		}
	}
	//Save existing settings
	GlobalObjects::Settings->Save();

	//Try to find the configurator
	String^ configuratorPath = Path::Combine(Path::GetDirectoryName(Settings::typeid->Assembly->Location), "Misnamed Filter\\Misnamed Filter.Configurator.exe");
	if(!File::Exists(configuratorPath))
	{
		//Configurator doesn't exist, so just launch the xml file in notepad
		System::Diagnostics::Process::Start("notepad.exe", Path::Combine(settingsFolder->FullName, "Misnamed Filter.settings.xml"))->WaitForExit();
	}
	else
	{
		//Run the configurator
		System::Diagnostics::Process::Start(configuratorPath)->WaitForExit();
	}
	//Reload the settings
	GlobalObjects::Settings->Reload();
}


String^ GetDisplayName(Match^ match)
{
	Group^ group = match->Groups["name"];
	if(group->Success)
	{
		return group->Value->TrimEnd(GlobalObjects::DisplayNameTrimChars)->TrimStart(GlobalObjects::DisplayNameTrimChars);			
	}
	return nullptr;
}

String^ DecodeHeaderValue(String^ value)
{
	//Call the internal CLR DecodeHeaderValue method.
	//This is not publically exposed, so have to guard against it disappearing in future CLR versions.
	if (GlobalObjects::DecodeHeaderValue != nullptr)
	{
		try
		{
			return (String^)GlobalObjects::DecodeHeaderValue->Invoke(nullptr, gcnew array<Object^> { value });
		}
		catch (Exception^)
		{
			//Couldn't decode the header (something went wrong)
		}
	}
	return value;
}

String^ GetAddress(Match^ match)
{
	return match->Groups["address"]->Value;
}

String^ GetHeaderBody(Match^ match)
{
	return match->Groups["header"]->Value;
}

/**
 * Check message by headers

 return value:  1 : Non-Spam   (no other filters will be applied)
                0 : unsure     (let other filters decide)
               -1 : Spam!
 */
int isSpamHdr(unsigned int id, const MSGHDR *header, double *probability)
{
	List<String^>^ misnamedAddresses = gcnew List<String^>(); //Keep a list of the misnamed addresses, to use as the spam words (reason)
	MSGHDRFIELD *temp = header->fieldlist;
	while (temp)
	{
		//Check the To, CC, and BCC headers
		if (temp->name && temp->body && 
				(stricmp(temp->name, "To") == 0 || stricmp(temp->name, "CC") == 0 || stricmp(temp->name, "BCC") == 0)
				)
		{
			for each (Match^ match in GlobalObjects::EmailParser->Matches(gcnew String(temp->body)))
			{
				String^ encodedDisplayName = GetDisplayName(match);
				
				if(!String::IsNullOrEmpty(encodedDisplayName)) //Only do any of the rest of the check if there is a display name to check
				{
					String^ displayName = DecodeHeaderValue(encodedDisplayName);
				
					//Get the email address
					String^ address = GetAddress(match);

					//Check if there is a rule for that email address, and if there is, if the display name is valid.
					bool addressMatched = false;
					bool foundGoodName = false;
					for each(AddressSet^ addressSet in GlobalObjects::Settings->AddressSets)
					{
						for each(String^ candidateAddressPattern in addressSet->AddressPatterns)
						{
							if(!String::IsNullOrEmpty(candidateAddressPattern))
							{
								try
								{
									//Treat the address as a regex and check it against the address
									if(Regex::IsMatch(address, candidateAddressPattern, RegexOptions::IgnoreCase | RegexOptions::ExplicitCapture))
									{
										addressMatched = true;
										//Found a matching address which needs checking, so check that the name is contained within the display name
										for each(String^ candidateNamePattern in addressSet->NamePatterns)
										{
											if(!String::IsNullOrEmpty(candidateNamePattern))
											{
												try
												{
													//Treat the name as a regex too, matching either encoded or decoded
													if( Regex::IsMatch(encodedDisplayName, candidateNamePattern, RegexOptions::IgnoreCase | RegexOptions::ExplicitCapture) ||
														Regex::IsMatch(displayName, candidateNamePattern, RegexOptions::IgnoreCase | RegexOptions::ExplicitCapture))
													{
														//This address is directed to a valid name.
														foundGoodName = true;
														break; //Don't need to keep looking through the other names, if any match, it's good for the set (OR)
													} // </Is the name a match>
												}
												catch (ArgumentException^)
												{
													//Ignore regex exception
												}
											} // </if not blank candidate name pattern>
										} // </For each namePattern>
										//Don't need to keep looking through the address set, if any address pattern matches, the set is treated as matching (OR)
										break; //Break out of for each candidateAddressPattern
									} // </Is the address a match>
								}
								catch (ArgumentException^)
								{
									//Ignore regex exception
								}
							} // </if not blank candidate address pattern>
							//If the address did not match, try the others in the set
						}// </For each candidateAddressPatern>
						if(foundGoodName) break; //If a good name has been found, then this address set is good.
						//If a good name wasn't found, then check the other address sets. A single address might match more than one set, and is OK if it has a valid name in any of them.
					} // </For each address set>
					
					if(addressMatched && !foundGoodName)
					{
						//At least one address matched, and there was no good name for it, so this is a misnamed address
						misnamedAddresses->Add(match->Value);
					}
				} // </Is DisplayName present>
			}// </For each email address>
		} // </Header is To, CC, or BCC>
		temp = temp->next;
	} // </Loop through all headers>

	if(misnamedAddresses->Count > 0)
	{
		//At least one address was misnamed, so this is spam, and should be reported
		String^ reason = String::Join(", ", misnamedAddresses->ToArray());
		IntPtr marshalledReason = System::Runtime::InteropServices::Marshal::StringToHGlobalAnsi(reason);
		
		//Report the tests as the Spamihilator 'spam words' (reason)
		SERVICE fsw;
		fsw.ulSize      = sizeof(fsw);
		fsw.strReceiver = RECV_SPAMIHILATOR;
		fsw.ulService   = SERV_FOUNDSPAMWORDS;
		fsw.wParam      = (short)id;
		fsw.lParam      = (long)(int)marshalledReason;
		if (SendService) SendService(&fsw);
		System::Runtime::InteropServices::Marshal::FreeHGlobal(marshalledReason);
				
		return -1; //Spam
	}
	else
	{
		//No misnamed addresses encountered, so can't say it is spam. Can't say it isn't, either, though.
		return 0; //Unsure
	}
	return 0;
}

/**
 * Learn from a message.
 *
 * @param flag Message flag. (SERVLEARNGOOD or SERVLEARNBAD)
 * @param messageText The whole message (new in 0.9.8: including headers and attachments)
 */
long onLearn(short flag, const char *messageText)
{
	if(flag == SERVLEARNGOOD && GlobalObjects::Settings->AutoLearn) //Can only learn from good messages
	{
		String^ message = gcnew String(messageText);
		int headersEndPos = message->IndexOf("\r\n\r\n"); //As mandated for email by rfc2822
		if(headersEndPos == -1)
			headersEndPos = message->IndexOf("\n\n"); //unix
		if(headersEndPos == -1)
			headersEndPos = message->IndexOf("\r\r"); //mac?
		
		if(headersEndPos > -1)
		{
			bool learnt = false;
			String^ headers = message->Substring(0, headersEndPos);
			
			for each(Match^ addressHeaderMatch in GlobalObjects::HeaderParser->Matches(headers))
			{
				String^ headerBody = GetHeaderBody(addressHeaderMatch);
				//Find addresses in this header
				for each(Match^ addressMatch in GlobalObjects::EmailParser->Matches(headerBody))
				{
					String^ encodedDisplayName = GetDisplayName(addressMatch);

					if(!String::IsNullOrEmpty(encodedDisplayName)) //Only do any of the rest of the check if there is a display name to learn from
					{
						String^ displayName = DecodeHeaderValue(encodedDisplayName);

						//Learn from this display name and address pair
						String^ address = GetAddress(addressMatch);

						//Find an existing address set which matches
						AddressSet^ addressSet = nullptr;
						for each(AddressSet^ candidateAddressSet in GlobalObjects::Settings->AddressSets)
						{
							for each(String^ candidateAddressPattern in candidateAddressSet->AddressPatterns)
							{
								if(!String::IsNullOrEmpty(candidateAddressPattern))
								{
									try
									{
										//Treat the address as a regex and check it against the address
										if(Regex::IsMatch(address, candidateAddressPattern, RegexOptions::IgnoreCase | RegexOptions::ExplicitCapture))
										{
											addressSet = candidateAddressSet;
											break; //Break out of for each candidateAddressPattern
										} // </Is the address a match>
									}
									catch (ArgumentException^)
									{
										//Ignore regex exception
									}
								}
								if(addressSet != nullptr) break;
								//If the address did not match, try the others in the set
							}// </For each candidateAddressPatern>
							if(addressSet != nullptr) break;
						} // </For each address set>

						if(addressSet == nullptr)
						{
							//Create a new address set, as no existing ones matched.
							addressSet = gcnew AddressSet();
							addressSet->DescriptiveName = String::Format("(auto) {0}", address);
							addressSet->AddressPatterns->Add(String::Format("^{0}$",Regex::Escape(address))); //Escape regex special characters, and anchor to start and end of string.
							GlobalObjects::Settings->AddressSets->Add(addressSet);
						}

						//Check if the name needs adding
						bool foundName = false;
						for each(String^ candidateNamePattern in addressSet->NamePatterns)
						{
							if(!String::IsNullOrEmpty(candidateNamePattern))
							{
								try
								{
									//Treat the name as a regex too, matching either encoded or decoded
									if( Regex::IsMatch(encodedDisplayName, candidateNamePattern, RegexOptions::IgnoreCase | RegexOptions::ExplicitCapture) ||
										Regex::IsMatch(displayName, candidateNamePattern, RegexOptions::IgnoreCase | RegexOptions::ExplicitCapture))
									{
										//There's already a matching name, so no need to learn further.
										foundName = true;
										break; //Don't need to keep looking through the other names, if any match, it's good for the set (OR)
									} // </Is the name a match>
								}
								catch (ArgumentException^)
								{
									//Ignore regex exception
								}
							}
						} // </For each namePattern>

						if(!foundName)
						{
							//Add the name
							addressSet->NamePatterns->Add(String::Format("^{0}$",Regex::Escape(displayName))); //Escape regex special characters, and anchor to start and end of string.
							learnt = true;
						}
					}
				}
			}
			if(learnt)
			{
				GlobalObjects::Settings->Save(); //Keep settings up to date
			}
			return TRUE;
		}
	}
	return FALSE;
}